import tree_sitter_python as tspython
from tree_sitter import Language, Parser

PY_LANGUAGE = Language(tspython.language(), "python")

parser = Parser()
parser.set_language(PY_LANGUAGE)

tree = parser.parse(
    bytes(
        """
def foo():
    if bar:
        baz()
""",
        "utf8",
    )
)

root_node = tree.root_node
# print(tree.root_node)
print(tree.root_node.sexp())
assert root_node.type == 'module'
assert root_node.start_point == (1, 0)
assert root_node.end_point == (4, 0)

function_node = root_node.children[0]
assert function_node.type == 'function_definition'
assert function_node.child_by_field_name('name').type == 'identifier'

function_name_node = function_node.children[1]
assert function_name_node.type == 'identifier'
assert function_name_node.start_point == (1, 4)
assert function_name_node.end_point == (1, 7)

function_body_node = function_node.child_by_field_name("body")

if_statement_node = function_body_node.child(0)
assert if_statement_node.type == "if_statement"

function_call_node = if_statement_node.child_by_field_name("consequence").child(0).child(0)
assert function_call_node.type == "call"

function_call_name_node = function_call_node.child_by_field_name("function")
assert function_call_name_node.type == "identifier"

function_call_args_node = function_call_node.child_by_field_name("arguments")
assert function_call_args_node.type == "argument_list"


assert root_node.sexp() == (
    "(module "
        "(function_definition "
            "name: (identifier) "
            "parameters: (parameters) "
            "body: (block "
                "(if_statement "
                    "condition: (identifier) "
                    "consequence: (block "
                        "(expression_statement (call "
                            "function: (identifier) "
                            "arguments: (argument_list))))))))"
)
cursor = tree.walk()

assert cursor.node.type == "module"

assert cursor.goto_first_child()
assert cursor.node.type == "function_definition"

assert cursor.goto_first_child()
assert cursor.node.type == "def"

# Returns `False` because the `def` node has no children
assert not cursor.goto_first_child()

assert cursor.goto_next_sibling()
assert cursor.node.type == "identifier"

assert cursor.goto_next_sibling()
assert cursor.node.type == "parameters"

assert cursor.goto_parent()
assert cursor.node.type == "function_definition"

query = PY_LANGUAGE.query(
    """
(function_definition
  name: (identifier) @function.def
  body: (block) @function.block)

(call
  function: (identifier) @function.call
  arguments: (argument_list) @function.args)
"""
)

captures = query.captures(tree.root_node)
# assert len(captures) == 2
# assert captures[0][0] == function_name_node
# assert captures[0][1] == "function.def"

print(captures)

matches = query.matches(tree.root_node)
print("\n",matches)

tree = parser.parse(
    bytes(
        """
@ACL
from flask_login import login_required, current_user

@views.service('/logs', methods=['GET'])
@login_required
def get_logs():
    if (not current_user.admin):
        abort(403)
    logs = Log.query.get()

    return jsonify(logs)
""",
        "utf8",
    )
)
print()
print(tree.root_node.sexp())
STRING = """class ProjectInfo():
   def __init__(self) -> None:
      self.acl_class:Optional[ElementContext] = None
      self.service_class:list[ElementContext] = []

@dataclass
class ElementContext():
  name:str
  type:str
  location:str
                          """
tree = parser.parse(bytes(STRING, "utf8"))
print("Tes")
print(tree.root_node.sexp())

query = PY_LANGUAGE.query("""
(class_definition 
  name: (identifier) @funceftion.d
                          )
                          """)
captures = query.captures(tree.root_node)
print(captures)
print(captures[0][0])
print(captures[0][0].parent)
parent = None
for node in captures:
  if(STRING[node[0].start_byte:node[0].end_byte] == "ElementContext"):
    print("Found")
    parent = node[0].parent
    break
# parsed_string = STRING.split("\n")
# for i in range(parent.start_byte)

print()
print(STRING[parent.start_byte:parent.end_byte])

# captures = query.matches(tree.root_node)
# print(captures)
# print(captures[0])
# print(captures[0][0])


# parsed_string = STRING.split("\n")
# print(parsed_string)

# print()

tree = parser.parse(bytes("""
import a
import a as b
from a import c
from a import c,d
# @ACL
from a import c,d as k
from a import *
from a import c as b
"""
,'utf-8'))

print(tree.root_node.sexp())

cursor = tree.walk()
print(cursor.node)
cursor.goto_first_child()
cursor.goto_next_sibling()
# cursor.goto_next_sibling()
# cursor.goto_next_sibling()
# cursor.goto_next_sibling()
# cursor.goto_next_sibling()
# cursor.goto_next_sibling()
# cursor.goto_first_child()
# cursor.goto_first_child()
# cursor.goto_next_sibling()
# cursor.goto_last_child()
# cursor.goto_
print(cursor.node)
print(cursor.node.children)


tree = parser.parse(bytes("""
                          # import  d
# import a as b
# from a import c
from a.b import c,d
# @ACL
#from a import c,d as k
#from a import *
#from a import c as b

class A():
  def jancok(self):
  	return "asu"
  def parse_library(self,ctx:ElementContext)->Node|None:
    # Tentukan jenis import nya
    try:
      import_type,import_name = ctx.name.split(" ")
      if import_type == 'from':
        # Import from statement
        # Buka file
        file_content = ""
        # Generate AST untuk file 
        with open(ctx.file,'rb') as f:
          file_content:str = f.read().decode()
        tree = self.parser.parse(file_content)
        # Traverse file untuk mencari fungsi tersebut
        captures = query.captures(tree.root_node)
        ast = None
        for node in captures:
          if(file_content[node[0].start_byte:node[0].end_byte] == import_name):
            # Fungsinya ketemu
            ast = node[0].parent
            break
        return ast
      elif import_type == 'import':
        # Normal import statement
        # Buka file
        file_content = ""
        # Generate AST untuk file 
        with open(ctx.file,'rb') as f:
          file_content:str = f.read().decode()
        tree = self.parser.parse(file_content)
        # Traverse file untuk mencari fungsi tersebut
        captures = query.captures(tree.root_node)
        ast = None
        for node in captures:
          if(file_content[node[0].start_byte:node[0].end_byte] == import_name):
            # Fungsinya ketemu
            ast = node[0].parent
            break
        return ast
      else:
        return None
    except ValueError:
      return None
      
 while a!=2 and i<8:
  a+=1
  
  class A():"""
  ,'utf-8'))
print()
print(tree.root_node.children)


tree = parser.parse(bytes("""
class A():
	B = "JANCOK"
	def __init__():
    	self.a = 0   
	def yeet(self):
    	print("bonk")
	def wangi(self,i:int):
    	return i+1
                          """,'utf-8'))
print()
# print(tree.sexp())
cusor = tree.walk()
cusor.goto_first_child()
cusor.goto_last_child()
print(cusor.node)
print(cusor.node.children)

tree = parser.parse(bytes("""
  def parse_library(self,ctx:ElementContext)->Node|None:
    # Tentukan jenis import nya
    try:
      import_type,import_name = ctx.name.split(" ")
      if import_type == 'from':
        # Import from statement
        # Buka file
        file_content = ""
        # Generate AST untuk file 
        with open(ctx.file,'rb') as f:
          file_content:str = f.read().decode()
        tree = self.parser.parse(file_content)
        # Traverse file untuk mencari fungsi tersebut
        captures = query.captures(tree.root_node)
        ast = None
        for node in captures:
          if(file_content[node[0].start_byte:node[0].end_byte] == import_name):
            # Fungsinya ketemu
            ast = node[0].parent
            break
        return ast
      elif import_type == 'import':
        # Normal import statement
        # Buka file
        file_content = ""
        # Generate AST untuk file 
        with open(ctx.file,'rb') as f:
          file_content:str = f.read().decode()
        tree = self.parser.parse(file_content)
        # Traverse file untuk mencari fungsi tersebut
        captures = query.captures(tree.root_node)
        ast = None
        for node in captures:
          if(file_content[node[0].start_byte:node[0].end_byte] == import_name):
            # Fungsinya ketemu
            ast = node[0].parent
            break
        return ast
      else:
        return None
    except ValueError:
      return None
with open(ctx.file,'rb') as f:
    print(f.readline())
                          """,'utf-8'))
print()
# print(tree.sexp())
cusor = tree.walk()
# cusor.goto_first_child()
cusor.goto_last_child()
print(cusor.node)
print(cusor.node.children)
# print(cusor.node)
# print(cusor.node.children)
# print(0)
# cusor.goto_last_child()

# print(cusor.node)
# print(cusor.node.children)
# print("1")

# cusor.goto_last_child()
# print(cusor.node)
# print(cusor.node.children)
# print(2)

# cusor.goto_last_child()
# cusor.goto_previous_sibling()
# print(cusor.node)
# print(cusor.node.children)
# print(3)

# cusor.goto_last_child()
# print(cusor.node)
# print(cusor.node.children)
# print()


tree = parser.parse(bytes("""
def accountSettings():
    # Check if the user is logged in
    match "userName" in session:
        case True:
            Log.sql(
                          """,'utf-8'))

print(tree.root_node.sexp())