from lib.FileReader import ProjectInfo,ElementContext
from lib.ACReader import ACLData
from lib.utils import format_log,flatten_node
from tree_sitter import Node
from .CFGGenerator import CFG
from .Const import PATH_SEPARATOR
class ACLAnalyzer():
  def __init__(self,project:ProjectInfo,acl_info:ACLData) -> None:
    self.project_info:ProjectInfo = project
    self.acl_info:ACLData = acl_info

  def analyze(self)->ACLData:
    real_acl:ACLData = ACLData(self.acl_info.principal_list,{},self.acl_info.acl_context)

    for route in self.project_info.service_class:
      format_log(f"Analyzing Service {route.get_base_element_name()} in {route.location} Control Access...")
      # Tentukan tipenya
      if(route.type == 'function'):
        service_acl = self.analyze_function(route)
        extra = ""
        for k in route.context:
          if k.startswith("Parent::"):
            extra = f"{k.split('::')[1]}."
        real_acl.service_acl[f"{extra}{route.get_base_element_name()}"] = service_acl
      elif(route.type=='class'):
        service_acls = self.analyze_class(route)
        for route in service_acls:
          real_acl.service_acl[route[0]] = route[1]
      elif(route.type=='module'):
        service_acls = self.analyze_module(route)
        for route in service_acls:
          real_acl.service_acl[route[0]] = route[1]
    return real_acl
  

  def analyze_class(self,route:ElementContext)->list[str]:
    format_log(f"Analyzing class {route.get_base_element_name()} in {route.location} Control Access...")
    file_method_list = self.project_info.dependency_manager.get(route.location).method_list
    method_lists = list(filter(lambda x:x['name']!='__init__' and x['type']=='class_method' and x['parent']==route.get_base_element_name(),file_method_list))
    # Remove yang ada di no check list
    for ctx in route.context:
      if ctx.startswith("Exempt::"):
        exempted = ctx.split("::")[1]
        for met in method_lists:
          if(met['name']==exempted):
            method_lists = [m for m in method_lists if m['name']!=met['name']]
    # Cek setiap fungsi
    class_acls = []
    for f in method_lists:
      ctx = ElementContext(f"{f['parent']}.{f['name']}",'function',route.location,[f"Parent::{f['parent']}"])
      ctx.set_cfg(CFG(f['ast'],f['ctx'],f['cursor'],route.cfg.source_code))
      result = self.analyze_function(ctx)
      class_acls.append([ctx.name,result])
    return class_acls

  def analyze_function(self,route:ElementContext,is_route=True)->list[str]:
    format_log(f"Analyzing function {route.get_base_element_name()} in {route.location} Control Access...")
    file_method_list = self.project_info.dependency_manager.get(route.location).method_list
    # Cek parent nya untuk menentukaan apakah ada dekorator
    # Inisialisasi principal_list
    var_list = []
    principal_list = [principal for principal in self.acl_info.principal_list]
    route.cfg.reset()
    node:Node = route.cfg.traverse()
    if(node.parent.type=='decorated_definition'):
      # Cari fungsi dekorator yang dipanggil
      for dec in node.parent.children:
        if dec.type=='decorator':
          if dec.children[-1].type=='call':
            dekorator = dec.children[-1].children[0]  
          else:
            dekorator = dec.children[-1]   
          # Cek apakah ada di fungsi atau library
          acl_result = None
          if is_route:
            acl_result = self.check_acl_list(route,dekorator.text.decode().split("(")[0].replace("@","").strip())
          if(acl_result!=None):
          # Ini memang fungsi acl
            principal_list = list(set(principal_list) & set(acl_result))
    # Gak ada dekorator/dekorator gak cocok
    # Traverse fungsinya
    while node:
      # Cek detail untuk beberapa tipe node
      if node.type in ['call','assignment']:
        # Cek call method
        # Format: [attribute,argument_list] | [identifier,argument_list]
        child_list = flatten_node(node.children)
        fun_name = ""
        var_name = ""
        for i in range(len(child_list)):
          if child_list[i] in [':',',',]:
            continue
          elif child_list[i] in [')']:
              fun_name = ""
          elif child_list[i] == '=':
              # Simpan nama variabel di left hand side
              var_name = fun_name
              var_list.append({'name':fun_name,'type':child_list[i+1]})
              fun_name = ""
          elif child_list[i]=='==':
            # Cek left side
            left_side = child_list[i-1]
            right_side = child_list[i+1]
          elif child_list[i] in ['(']:
            # Cek apakah fungsinya ada
            if fun_name:
              # cek buat fungsi di kelas yang sama
              for rute in file_method_list:
                function_name = ""
                if fun_name.split(".")[0]!='self' and rute['type']=='module_function' and rute['name']==fun_name:
                  function_name = fun_name
                elif rute['type']=='class_method' and rute['name']==fun_name.split(".")[-1]:
                  if fun_name.split(".")[0]!='self':
                    function_name = fun_name
                  else:
                    function_name = fun_name.split(".")[-1]
                else:
                  function_name = fun_name
                # Cek
                acl_result = None
                if is_route:
                  acl_result =  self.check_acl_list(route,function_name)
                if acl_result!=None:
                  # Ada pengecekan
                  principal_list = list(set(principal_list) & set(acl_result))
              fun_name = ""
          else:
            fun_name += child_list[i]
        if var_name:
          try:
            acl_result = self.acl_info.acl_context[var_name]
            var_name = ""
          except KeyError:
            pass
      elif node.type=='comparison_operator':
        kinder = node.children
        var_list = []
        for i in range(len(kinder)):
          if kinder[i].type in ['call','attribute']:
            var_list = flatten_node(kinder[i].children)
          elif kinder[i].type in ["==","!="]:
            var_list.append(kinder[i].type)
          elif kinder[i].type=='string':
              var_list.append(kinder[i].text.decode())
        mode = ''
        role_name = ""
        var_key = ""
        for var in var_list:
          if var=='==':
            mode = 'eq'
          elif var=='!=':
            mode = 'neq'
          elif "\"" in var or '\'' in var:
            # String
            # Cek apakah acl nya dict
            role_name = var.replace("\"","").replace("\'","")
          else:
            # Cek key
            var_key = "::".join([f"var:{c}" for c in var.split(".")])
          if var_key in self.acl_info.acl_context and mode and role_name:
              if type(self.acl_info.acl_context[var_key])==list:
                pass
              elif type(self.acl_info.acl_context[var_key])==dict:
                # Cek mode
                if mode == 'eq':
                  acl_result = [self.acl_info.acl_context[var_key][role_name]] if role_name in self.acl_info.acl_context[var_key] else []
                  principal_list = list(set(principal_list) & set(acl_result))
                elif mode=='neq':
                  acl_result = []
                  for role in self.acl_info.acl_context[var_key]:
                    if role!=role_name:
                      acl_result.append(self.acl_info.acl_context[var_key][role])
                  principal_list = list(set(principal_list) & set(acl_result))
          # elif kinder
      elif node.type=='while_statement':
        for elmt in node.children:
          # Skip kalau dah ':'
          if elmt.type==':':
            break
          elif elmt.type=='while':
            continue
          else:
            # Baca kondisi
            child_list = flatten_node(node.children)
            fun_name = ""
            for child in child_list:
              if child in [':',',']:
                continue
              elif child in ['=',')']:
                fun_name = ""
              elif child in ['('] and fun_name:
                # Cek apakah fungsinya ada
                if fun_name:
                  # cek buat fungsi di kelas yang sama
                  for rute in file_method_list:
                    function_name = ""
                    if fun_name.split(".")[0]!='self' and rute['type']=='module_function' and rute['name']==fun_name:
                      function_name = fun_name
                    elif rute['type']=='class_method' and rute['name']==fun_name.split(".")[-1]:
                      if fun_name.split(".")[0]!='self':
                        function_name = fun_name
                      else:
                        function_name = fun_name.split(".")[-1]
                    # Cek
                    acl_result = None
                    if is_route:
                      acl_result =  self.check_acl_list(route,function_name)
                    if acl_result!=None:
                      # Ada pengecekan
                      principal_list = list(set(principal_list) & set(acl_result))
                  fun_name = ""
              else:
                fun_name += child
      elif node.type=='for_statement':
        # Cek kondisinya dulu
        components = []
        for child in node.children:
          if child.type in ['for',':','in','block']:
            continue
          elif child.type=='identifier':
            components.append(child.text.decode())
          else:
            components += flatten_node(child.children)
        # Cek tiap componentnya
        child_list = flatten_node(node.children)
        fun_name = ""
        for child in child_list:
          if child in [':',',']:
            continue
          elif child in ['=',')']:
            fun_name = ""
          elif child in ['('] and fun_name:
            # Cek apakah fungsinya ada
            if fun_name:
              # cek buat fungsi di kelas yang sama
              for rute in file_method_list:
                function_name = ""
                if fun_name.split(".")[0]!='self' and rute['type']=='module_function' and rute['name']==fun_name:
                  function_name = fun_name
                elif rute['type']=='class_method' and rute['name']==fun_name.split(".")[-1]:
                  if fun_name.split(".")[0]!='self':
                    function_name = fun_name
                  else:
                    function_name = fun_name.split(".")[-1]
                acl_result = None
                if is_route:
                  acl_result =  self.check_acl_list(route,function_name)
                if acl_result!=None:
                  # Ada pengecekan
                  principal_list = list(set(principal_list) & set(acl_result))
              fun_name = ""
          else:
            fun_name += child
      elif node.type=='with_statement':
        # Cek kondisinya dulu
        components = []
        for child in node.children:
          if child.type in ['with','block',":"]:
            continue
          elif child.type=='identifier':
            components.append(child.text.decode())
          else:
            components += flatten_node(child.children)
        # Cek tiap componentnya
        fun_name = ""
        for i in range(len(components)):
          if components[i] in [':',',']:
            continue
          elif child in ['=',')']:
            fun_name = ""
          elif components[i] in ['(']:
            # Cek apakah fungsinya dipanggil
            if fun_name:
              # cek buat fungsi di kelas yang sama
              for rute in file_method_list:
                function_name = ""
                if fun_name.split(".")[0]!='self' and rute['type']=='module_function' and rute['name']==fun_name:
                  function_name = fun_name
                elif rute['type']=='class_method' and rute['name']==fun_name.split(".")[-1]:
                  if fun_name.split(".")[0]!='self':
                    function_name = fun_name
                  else:
                    function_name = fun_name.split(".")[-1]
                # Cek
                acl_result = None
                if is_route:
                  acl_result =  self.check_acl_list(route,function_name)
                if acl_result!=None:
                  # Ada pengecekan
                  principal_list = list(set(principal_list) & set(acl_result))
              fun_name = ""
            elif components[i]=='as':
              prefix = components[i+1]
          else:
            # Tambahkan nama fungsinya
            fun_name += components[i]
      elif node.type in ['try_statement','finally_clause','else_clause']:
        pass
      elif node.type=='except_clause':
        # Dapatkan identifier
        identifier = node.children[1]
      elif node.type in ['if_statement','elif_clause']:
        # Cek kondisinya dulu
        components = []
        for child in node.children:
          if child.type in ['elif',':','block']:
            continue
          elif child.type=='identifier':
            components.append(child.text.decode())
          else:
            components += flatten_node(child.children)
        # Cek tiap componentnya
        fun_name = ""
        for i in range(len(components)):
          if components[i] in ['and','or','not']:
            continue
          elif components[i] in ['(',')']:
            # Cek apakah fungsinya dipanggil
            if fun_name:
              parts = fun_name.split(".")
              # Coba konversi dulu
              if len(parts)>1:
                for var in var_list:
                  if parts[0]==var['name']:
                    parts[0] = var['type']
                fun_name = ".".join(parts)    
              for rute in file_method_list:
                function_name = ""
                if parts[0]!='self' and rute['type']=='module_function' and rute['name']==fun_name:
                  function_name = fun_name
                elif rute['type']=='class_method' and rute['name']==parts[-1]:
                  if parts[0]!='self':
                    function_name = fun_name
                  else:
                    function_name = parts[-1]
                else:
                  function_name = fun_name
                # Cek
                acl_result = None
                if is_route:
                  acl_result =  self.check_acl_list(route,function_name,components[i+1:])
                if acl_result!=None:
                  # Ada pengecekan
                  principal_list = list(set(principal_list) & set(acl_result))
              fun_name = ""
          elif components[i] in ['==','!=']:
            # Cek apakah variabel sebelum atau sesudanya adalah acl context
            left_var_name = self.search_acl_var(fun_name,route)
            right_var_name = self.search_acl_var(components[i+1],route)
            if right_var_name in self.acl_info.acl_context:
              # Sanitasi left var name
              left_var_name = left_var_name.replace('\"',"").replace("\'","")
              components[i-1] = components[i-1].replace('\"',"").replace("\'","")
              # Cek apakah right var name adalah string
              if type(self.acl_info.acl_context[right_var_name])==dict:
                # Cek nilainya
                if components[i-1] in self.acl_info.acl_context[left_var_name]:
                  if components[i]=='==':
                    principal_list = list(set(principal_list) & set([self.acl_info.acl_context[left_var_name][components[i-1]]]))
                  else:
                    principal_list = list(set(principal_list) & set([x for x in self.acl_info.principal_list if x!=self.acl_info.acl_context[left_var_name][components[i-1]]]))
              else:
                # Langsung cari yg terkecil
                principal_list = list(set(principal_list) & set(self.acl_info.acl_context[right_var_name]))
            elif left_var_name in self.acl_info.acl_context:
              # Sanitasi
              right_var_name = right_var_name.replace('\"',"").replace("\'","")
              components[i+1] = components[i+1].replace('\"',"").replace("\'","")
              # Cek apakah right var name adalah string
              if type(self.acl_info.acl_context[left_var_name])==dict:
                # Cek nilainya
                if components[i+1] in self.acl_info.acl_context[left_var_name]:
                  if components[i]=='==':
                    principal_list = list(set(principal_list) & set([self.acl_info.acl_context[left_var_name][components[i+1]]]))
                  else:
                    principal_list = list(set(principal_list) & set([x for x in self.acl_info.principal_list if x!=self.acl_info.acl_context[left_var_name][components[i+1]]]))
              else:
                # Langsung cari yg terkecil
                principal_list = list(set(principal_list) & set(self.acl_info.acl_context[left_var_name]))
          elif components[i] in ['in']:
            # Cek apakah variaabel kiri merupakan key yang di kanan
            fun_name = fun_name.replace('\"',"").replace("\'","")
            acl_key = self.search_acl_var(f"{components[i+1]}.{fun_name}",route)
            if acl_key:
              #Langsung cari yg terkecil
              principal_list = list(set(principal_list) & set(self.acl_info.acl_context[acl_key]))
          elif components[i] in ['>','<','>=','<=']:
            continue
          else:
            # Tambahkan nama fungsinya
            fun_name += components[i]
      node = route.cfg.traverse()
    return principal_list

  def analyze_module(self,route:ElementContext)->list[list[str,list[str]]]:
    format_log(f"Analyzing module {route.get_base_element_name()} in {route.location} Control Access...")
    file_method_list = self.project_info.dependency_manager.get(route.location).method_list
    # Cari info semua fungsi di module
    method_lists = [m for m in file_method_list if not (m['type']=='class_method' and m['name']=='__init__')]
    # Remove yang ada di no check list
    for ctx in route.context:
      if ctx.startswith("Exempt::"):
        exempted = ctx.split("::")[1]
        for met in method_lists:
          if(met['name']==exempted or (met['type']=='class_method' and met['parent']==exempted)):
            method_lists = [m for m in method_lists if m['name']!=met['name']]
    # Cek setiap fungsi
    service_acls = []
    for f in method_lists:
      if f['type']=='class_method':
        ctx = ElementContext(f"{f['parent']}.{f['name']}",'function',route.location,[])
        ctx.set_cfg(CFG(f['ast'],f['ctx'],f['cursor'],route.cfg.source_code))
        result = self.analyze_function(ctx)
        service_acls.append([ctx.get_base_element_name(),result])
      else:
        ctx = ElementContext(f['name'],'function',route.location,[])
        ctx.set_cfg(CFG(f['ast'],f['ctx'],f['cursor'],route.cfg.source_code))
        result = self.analyze_function(ctx)
        service_acls.append([ctx.get_base_element_name(),result])
    return service_acls
  
  def check_acl_list(self,route:ElementContext,name:str,ctx:list[str]=[])->list[str]|None:
    file_method_list = self.project_info.dependency_manager.get(route.location).method_list
    # Format name: fun atau self.fun
    acl_list = [role for role in self.acl_info.principal_list]
    for acl_class in self.project_info.acl_class:
      acl_class.cfg.reset()
      if(acl_class.type=='library' and acl_class.location==route.location):
        if(acl_class.context):
          # Cek apakah fungsi yang merupakan acl digunakan
          # context berisi daftar fungsi yang merupakan fungsi untuk cek ACL
          while True:
            acl_node:Node|None = acl_class.cfg.traverse()
            if not acl_node:
              break
            # Handle kasus antara gak ada context fungsi mana yang acl maupun ada
            if(not acl_class.context or  (acl_node.text.decode().split(".")[0] in acl_class.context)):
              # Cek jika ini fungsi ataupun kelas
              if(name==acl_class.cfg.get_name(acl_node) or name.split(".")[0]==acl_class.cfg.get_name(acl_node)):
              # Cek apakah dia manggil fungsi yang diimport di acl
                elmt = acl_class.get_base_element_name().split(" ")
                formatted_key = f"lib:{elmt[-1]}::{name}"
                if formatted_key not in self.acl_info.acl_context:
                  acl_list = None
                else:
                  acl_list = self.acl_info.acl_context[formatted_key]
      # Kasus kalau acl nya kelas atau fungsi
      elif acl_class.type in ['class','function']:
        # Cek apakah ada parent context
        parent_class = ""
        for k in acl_class.context:
          if k.startswith("Parent::"):
            parent_class = k.split("::")[1]
        if acl_class.location==route.location:
          # Kelas/fungsi lokal
          # Kali aja ada di module ini
          for function in file_method_list:
            if(acl_class.type=='class' and function['type']=='class_method' and function['parent']==acl_class.get_base_element_name()):
              full_name = name
              if len(name.split("."))==1 and ctx:
                j = 0
                elmt = ctx[j]
                while elmt in ['(',')',',']:
                  j+=1
                  elmt = ctx[j]
                full_name = f"{name}.{elmt}"
              if(full_name==f"{function['parent']}.{function['name']}"):
                # Cek apakah dia manggil fungsi yang diimport di acl
                this_acl = None
                for acl in self.analyze_class(acl_class):
                  if acl[0]==full_name:
                    this_acl = acl
                    break
                if this_acl:
                  acl_list = list(set(acl_list) & set(acl[1]))
            elif(acl_class.type=='function'):
              if(function['type']=='module_function' and function['name']==acl_class.get_base_element_name() and name==function['name']):
              # Cek si cfg kelas acl
                acl_list = list(set(acl_list) & set(self.analyze_function(acl_class,False)))
              elif(function['type']=='class_method' and function['name']==acl_class.get_base_element_name() and (name==f"{function['parent']}.{function['name']}" or function['parent']==parent_class)):
                acl_list = list(set(acl_list) & set(self.analyze_function(acl_class,False)))
        else:
          # Beda file, cek di import ada gak
          # Kali aja ada di module ini
          dependency_lists = self.project_info.dependency_manager.get(route.location).dependency_list
          route_position =  route.cfg.get_line_index(route.name)
          for dependency in dependency_lists:
            if dependency[0 ]> route_position:
              break
            dependency_parts = dependency[1]['original'] # Module name
            if dependency_parts not in acl_class.location.replace(PATH_SEPARATOR,"."):
              # Modul nya salah
              continue
            # Yang diimport class nya
            if(acl_class.type=='class'):
              # Cukup cek apakah kelas atau fungsinya sama
                for m in dependency[2]:
                  comparator = m['original']
                  if m["rename"]:
                    # Pake rename
                    comparator = m["rename"]
                  if len(name.split("."))==1 and ctx:
                    j = 0
                    elmt = ctx[j]
                    j+=1
                    while j < len(ctx) and elmt in ['(',')',',']:
                      elmt = ctx[j]
                      j+=1
                    name = f"{name}.{elmt}"
                  acl_file_method_list = self.project_info.dependency_manager.get(acl_class.location).method_list
                  for method in acl_file_method_list:
                    if method['type']!='class_method' or (method['type']=='class_method' and method['parent']!=m['original']):
                      continue
                    if(name==f'{comparator}.{method["name"]}'):
                      # Cek apakah dia manggil fungsi yang diimport di acl
                      this_acl = None
                      for acl in self.analyze_class(acl_class):
                        if acl[0]==name:
                          this_acl = acl
                          break
                      if this_acl:
                        acl_list = list(set(acl_list) & set(acl[1]))
            elif(acl_class.type=='function'):
              # Cukup cek apakah kelas atau fungsinya sama
                for m in dependency[2]:
                  comparator = m['original']
                  if m["rename"]:
                    # Pake rename
                    comparator = m["rename"]
                  tmp_name = name
                  if parent_class:
                    tmp_name = f"{parent_class}.{name}"
                  acl_file_method_list = self.project_info.dependency_manager.get(acl_class.location).method_list
                  for method in acl_file_method_list:
                    tmp_comparator = comparator
                    # Fungsi tidal sesuai
                    if method['type']=='module_function' and method['name']!=m['original']:
                      continue
                    elif method['type']=='class_method':
                      tmp_comparator = f"{tmp_comparator}.{method['name']}"
                    if(tmp_name==tmp_comparator):
                      # Cek apakah dia manggil fungsi yang diimport di acl
                      acl_list = list(set(acl_list) & set(self.analyze_function(acl_class,False)))
    return acl_list
  
  def search_acl_var(self,var_name:str,route:ElementContext)->str:
    # Jika ini variabel
    components = var_name.split(".")
    check_var = "::".join([f"var:{var}" for var in components])
    if check_var in self.acl_info.acl_context:
      return check_var
    # Jika ini library
    for dep in self.project_info.dependency_manager.get(route.location).dependency_list:
      lib_name = dep[1]['original']
      format_lib = f'lib:{lib_name}::{check_var}'
      if (components[0]==dep[1]['rename'] or components[0]==dep[1]['original']) and format_lib in self.acl_info.acl_context:
        return format_lib
      elif format_lib in self.acl_info.acl_context:
        return format_lib
      else:
        # Kasus gak ada namespace
        for method in dep[2]:
          if (method['rename'] and method['rename']==components[0]) or (not method['rename'] and method['original']==components[0]):
            format_lib = f"lib:{dep[1]['original']}::{method['original']}"
            return format_lib
    return ""