from lib.utils import format_log
import re

class ACLData():
  def __init__(self,principal_list:list[str],service_acl:dict[str,list[str]],acl_context:tuple[dict[str,list|dict]]) -> None:
    self.principal_list:list[str] = principal_list
    self.service_acl:dict[str,list[str]] = service_acl
    self.acl_context:tuple[dict[str,list|dict]] = acl_context

  def __str__(self):
    return f"ACLData(principals={self.principal_list},acl={self.service_acl},context={self.acl_context})"
  def __repr__(self):
    return self.__str__()
  

class ACReader():
  def __init__(self,path:str,req_file:str|None=None) -> None:
    self.acl_path:str = path
    self.req_file:str = req_file
  def read(self) -> ACLData:
    format_log("Reading ACL File...")
    principal_list:list[str] = []
    service_acl:dict[str,list[str]] = {}
    acl_context:dict[str,list|dict] = {}
    # Baca File ACL
    with open(self.acl_path) as acl_file:
      line_type = None
      for line in acl_file:
        # Handle Principal List
        if "=====PRINCIPAL LIST=====" in line:
          # Baca daftar principal
          line_type = 'principal'
        elif "=====ROUTES=====" in line:
          # Baca daftar routes
          line_type = 'routes'
        elif "=====DETAIL=====" in line:
          line_type = 'detail'
          # Baca detail
        else:
          # Parse sesuai tipe
          if line_type=='principal':
            principal_list = line.strip().replace(" ","").split(",")
          elif line_type=='routes':
            processed_line = line.strip().replace(" ","")
            if not processed_line:
              continue
            route = processed_line.split(":")
            if len(route)!=2:
              # Harus dalam format route:principal
              raise ValueError("Invalid ACL Data")
            # kasus ada penanda seluruh principal boleh (*)
            is_all = False
            for p in route[1]:
              if "*" in p:
                is_all = True
                break
            if is_all:
              service_acl[route[0]] = [principal for principal in principal_list]
            else:
              service_acl[route[0]] = route[1].split(",")
          elif line_type=='detail':
            # Parse detail
            processed_line = line.strip().replace(" ","")
            if not processed_line:
              continue
            detail = processed_line.split(";")
            if len(detail)!=2:
              # Harus dalam format var:context::var:var_name; principal
              raise ValueError("Invalid ACL Data")
            content = None
            bracket = re.search(r"\[.*\]",detail[1])
            if bracket:
              # Ada bracket, bikin dict konversi values dan principal
              bracket_contents = re.sub(r"[\[\]]","",bracket.group(0)).split(",")      
              content = {}
              for value in bracket_contents:
                subvalue = value.replace(" ","").split(":")
                if len(subvalue)!=2:
                  raise ValueError("Invalid ACL Data")
                content[subvalue[0]] = subvalue[1]
            else:
              content = detail[1].replace(" ","").split(",")
            acl_context[detail[0]] = content
    acl_data = ACLData(principal_list,service_acl,acl_context)
    return acl_data