import os
import re
from lib.utils import format_log
from lib.CFGGenerator import CFGGenerator,CFG
from lib.custom_class import ElementContext
from .Const import PATH_SEPARATOR
from .DependencyManager import DependencyManager
class FileReader:
    ANNOTATION_PATTERN = "\S*@.+"
    def __init__(self, project_path:str)->None:
        format_log("Initializing Project Context Gatherer...")
        self.project_path = project_path
        self.supported_extensions = os.getenv("SUPPORTED_EXTENSION").split("|")
        self.cfg_generator:CFGGenerator = CFGGenerator()
        self.dependency_names = []
        format_log("Project Context Gatherer initialized.")

    def analyze_project(self):
      format_log("Searching for annotation...")
      # Baca requirements.txt
      self.read_project_dependency()
      # Baca files
      potential_files:list[str] = self.get_potential_files()
      project_information:ProjectInfo = self.search_annotation(potential_files)
      return project_information
    
    def read_project_dependency(self):
      try:
        with open(f"{self.project_path}{PATH_SEPARATOR}requirements.txt") as req_file:
          lines = req_file.readlines()
          i = 0
          while i < len(lines):
            annotation_list = self.have_annotation(lines[i])
            annotation = None if len(annotation_list)==0 else annotation_list[0]
            if(annotation):
              additional_context = []
              if("@ACL" in annotation):
                # Ini kelas acl
                # Cek jika ada keterangan tambahan
                match = re.search(r'\(.*\)',lines[i])
                if match:
                  # Ada keterangan tambahan
                  additional_context = re.sub(r"[\(\) ]","",match.group(0)).split(",")
                # Cari tahu jenis acl nya
                i+=1
                while re.search(FileReader.ANNOTATION_PATTERN,lines[i]):
                  i+=1
                # Baca nama modul
                # import_name:flask_login
                module_name = additional_context[0].split(":")[1].strip()
                self.dependency_names.append({"name":module_name,"ctx":additional_context})
            i+=1
      except FileNotFoundError:
        format_log("requirements.txt not found. Skipping dependency reading...")

    def search_annotation(self,files:list[str]):
      project_info:ProjectInfo = ProjectInfo()
      for file in files:
        with open(file,'rb') as source_code:
          lines = source_code.read().decode().split("\n")
          i = 0
          while i < len(lines):
            is_skipped = False
            annotation_list = self.have_annotation(lines[i])
            annotation = None if len(annotation_list)==0 else annotation_list[0]
            if(annotation):
              additional_context = []
              if("@NoCheck" in annotation):
                # skip fungsi ini
                # Cari tahu jenis yang di-excempt
                i+=1
                stripped_lines = lines[i].strip()
                while not is_skipped and (re.search(FileReader.ANNOTATION_PATTERN,lines[i]) or not stripped_lines or stripped_lines[0] in ["#","@"]):
                  next_annotations = self.have_annotation(lines[i])
                  for annon in next_annotations:
                    if "@Service" in annon:
                      # Skip fungsi ini
                      is_skipped = True
                      break
                  i+=1
                  stripped_lines = lines[i].strip()
                if is_skipped:
                  continue
                # Kalau next line nya class, ya class, kalau def yg fungsi, selain itu type nya module
                curr_node = None
                if re.search("\S*class +",lines[i]): 
                  curr_node = ElementContext(lines[i].strip().split(" ")[1],'class',file,additional_context)
                elif re.search("\S*def +",lines[i]):
                  curr_node = ElementContext(lines[i].strip().split(" ")[1],'function',file,additional_context)
                else:
                  continue
                # Tentukan parent node ini
                tmp,_= self.cfg_generator.generate(curr_node)
                parent = tmp.ast.parent
                if parent.type=='decorated_definition':
                  parent = parent.parent
                if parent.type=='block':
                  parent = parent.parent
                # Kasih keterangan tambahan ke konteks modul/fungsi yang merupakan parent fungsi ini.
                for route in project_info.service_class:
                  tmp_parent = parent
                  if tmp_parent.type=='class_definition' and route.type=='module':
                    while tmp_parent.type != 'module':
                      tmp_parent = tmp_parent.parent
                  if route.location==file and (CFG.is_node_equal(route.cfg.ast,tmp_parent)):
                    route.context.append(f"Exempt::{curr_node.name.split('(')[0]}")
              elif("@ACL" in annotation):
                # Ini kelas acl
                # Cek jika ada keterangan tambahan
                match = re.search(r'\(.*\)',lines[i])
                if match:
                  # Ada keterangan tambahan
                  additional_context = re.sub(r"[\(\) ]","",match.group(0)).split(",")
                # Cari tahu jenis acl nya
                i+=1
                while re.search(FileReader.ANNOTATION_PATTERN,lines[i]):
                  i+=1
                if re.search("\S*class +",lines[i]):
                  # ACL nya kelas
                  project_info.acl_class.append(ElementContext(lines[i].strip().split(" ")[1],'class',file,additional_context))
                elif re.search("\S*def +",lines[i]):
                  # ACL nya fungsi
                  project_info.acl_class.append(ElementContext(lines[i].strip().split(" ")[1],'function',file,additional_context))
                elif re.search("\S*(from|import) +",lines[i]):
                  # ACL nya library eksternal
                  project_info.acl_class.append(ElementContext(lines[i],'library',file,additional_context))
                else:
                  # Ini kelas biasa
                  continue
                # Generate CFG jika ini acl
                cfg,deps = self.cfg_generator.generate(project_info.acl_class[-1])
                project_info.acl_class[-1].set_cfg(cfg)
                # Daftarkan dependency
                project_info.dependency_manager.add(deps)
                # Cek jika ini fungsi dalam kelas
                if(project_info.acl_class[-1].type=='function'):
                  parent = project_info.acl_class[-1].cfg.ast.parent.parent
                  if project_info.acl_class[-1].cfg.ast.parent.type=='decorated_definition':
                    parent = parent.parent
                  if parent and parent.type=="class_definition":
                      project_info.acl_class[-1].context.append(f"Parent::{parent.children[1].text.decode()}")
              elif("@Service" in annotation):
                # Ini kelas routes
                # Cek jika ada keterangan tambahan
                match = re.search(r'\((.*?)\)',lines[i])
                if match:
                  # Ada keterangan tambahan
                  additional_context = re.sub(r"[\(\) ]","",match.group(0)).split(",")
                # Cari tahu jenis rute nya
                i+=1
                have_annotation = self.have_annotation(lines[i])
                while not is_skipped and re.search(FileReader.ANNOTATION_PATTERN,lines[i]) and not have_annotation:
                  next_annotations = self.have_annotation(lines[i])
                  for annon in next_annotations:
                    if "@NoCheck" in annon:
                      # Skip fungsi ini
                      is_skipped = True
                      break
                  i+=1
                if have_annotation:
                  subanot = None if len(have_annotation)==0 else have_annotation[0]
                  if "@ACL" in subanot:
                    # Ini berarti fungsi ini buat module
                    project_info.service_class.append(ElementContext(".".join(file.split(PATH_SEPARATOR)[-1].split(".")[:-1]),'module',file,additional_context))
                    # Generate CFG jika ini rute
                    cfg,deps = self.cfg_generator.generate(project_info.service_class[-1])
                    project_info.service_class[-1].set_cfg(cfg)
                    project_info.dependency_manager.add(deps)
                    # Cek jika ini fungsi dalam kelas
                    if(project_info.service_class[-1].type=='function'):
                      parent = project_info.service_class[-1].cfg.ast.parent.parent
                      if project_info.service_class[-1].cfg.ast.parent.type=='decorated_definition':
                        parent = parent.parent
                      if parent and parent.type=="class_definition":
                          project_info.service_class[-1].context.append(f"Parent::{parent.children[1].text.decode()}")
                    # Langsung lanjut ke selanjutnya
                    continue
                if is_skipped:
                  continue
                # Kalau next line nya class, ya class, kalau def yg fungsi, selain itu type nya module
                if re.search("\S*class +",lines[i]): 
                  project_info.service_class.append(ElementContext(lines[i].strip().split(" ")[1],'class',file,additional_context))
                elif re.search("\S*def +",lines[i]):
                  project_info.service_class.append(ElementContext(lines[i].strip().split(" ")[1],'function',file,additional_context))
                else:
                  # Dapatkan konteks berupa cfg setiap kelas
                  project_info.service_class.append(ElementContext(".".join(file.split(PATH_SEPARATOR)[-1].split(".")[:-1]),'module',file,additional_context))
                # Generate CFG jika ini rute
                cfg,deps = self.cfg_generator.generate(project_info.service_class[-1])
                project_info.service_class[-1].set_cfg(cfg)
                project_info.dependency_manager.add(deps)
                # Cek jika ini fungsi dalam kelas
                if(project_info.service_class[-1].type=='function'):
                  parent = project_info.service_class[-1].cfg.ast.parent.parent
                  if project_info.service_class[-1].cfg.ast.parent.type=='decorated_definition':
                    parent = parent.parent
                  if parent and parent.type=="class_definition":
                      project_info.service_class[-1].context.append(f"Parent::{parent.children[1].text.decode()}")
              else:
                # Ini kelas biasa
                pass
            elif re.search("\S*(from|import) +",lines[i]):
              # ini modul yg diimport dan didefinisikan di requirement.txt
              for lib in self.dependency_names:
                if lib['name']==lines[i].split(" ")[1]:
                  # Bikin elemen
                  project_info.acl_class.append(ElementContext(lines[i],'library',file,lib['ctx']))
                  cfg,deps = self.cfg_generator.generate(project_info.acl_class[-1])
                  project_info.acl_class[-1].set_cfg(cfg)
                  # Daftarkan dependency
                  project_info.dependency_manager.add(deps)
            i+=1
      return project_info

    def have_annotation(self,line:str)->list[str]:
      return re.findall(r" *# *@[A-Za-z]+",line)
    
    def get_potential_files(self)->list[str]:
      file_list:list[str] = []
      for root, dirs, files in os.walk(self.project_path):
        for file in files:
          if(self.is_source_code_supported(file)):
            file_list.append(f"{root}{PATH_SEPARATOR}{file}")
      return file_list
    
    def is_source_code_supported(self,filename:str)->bool:
      extension = filename.split(".")[-1].lower()
      return extension in self.supported_extensions
    

class ProjectInfo():
  def __init__(self) -> None:
      self.acl_class:list[ElementContext] = []
      self.service_class:list[ElementContext] = []
      self.dependency_manager:DependencyManager= DependencyManager()
  
  def __repr__(self) -> str:
    return self.__str__()
  
  def __str__(self) -> str:
    return f"ProjectInfo(acl_class={self.acl_class},service_class={self.service_class},dependency_manager={self.dependency_manager})"
  