from lib.FileReader import ProjectInfo
from lib.custom_class import ElementContext
from lib.utils import format_log,flatten_node
from .CFGGenerator import CFG
from tree_sitter import Node
from .Const import PATH_SEPARATOR
class HandlerSanitizationAnalyzer():
  def set_project(self,project_info:ProjectInfo):
    self.project_info = project_info
    return self
  def analyze(self)->list[str]:
    vuln_routes = []
    for route in self.project_info.service_class:
      format_log(f"Analyzing Service {route.get_base_element_name()} in {route.location}...")
      # Tentukan tipenya
      if(route.type == 'function'):
        is_sanitized = self.analyze_function(route)
        if not is_sanitized:
          extra = ""
          for k in route.context:
            if k.startswith("Parent::"):
              extra = f"{k.split('::')[1]}."
          vuln_routes.append(f"{extra}{route.get_base_element_name()}() in {route.location}")
      elif(route.type=='class'):
        unsanitized_functions = self.analyze_class(route)
        for func in unsanitized_functions:
          vuln_routes.append(f"{route.get_base_element_name()}.{func}() in {route.location}")
      elif(route.type=='module'):
        unsanitized_functions = self.analyze_module(route)
        for func in unsanitized_functions:
          if func[0]=='function':
            vuln_routes.append(f"{func[1]}() in {route.location}")
          else:
            vuln_routes.append(f"{func[2]}.{func[1]}() in {route.location}")
    return vuln_routes

  def analyze_class(self,route:ElementContext)->list[str]:
    format_log(f"Analyzing class {route.get_base_element_name()} in {route.location}...")
    file_method_list = self.project_info.dependency_manager.get(route.location).method_list
    method_lists = list(filter(lambda x:x['name']!='__init__' and x['type']=='class_method' and x['parent']==route.get_base_element_name(),file_method_list))
    # Remove yang ada di no check list
    for ctx in route.context:
      if ctx.startswith("Exempt::"):
        exempted = ctx.split("::")[1]
        for met in method_lists:
          if(met['name']==exempted):
            method_lists = [m for m in method_lists if m['name']!=met['name']]
    # Cek setiap fungsi
    unsanitized_methods = []
    for f in method_lists:
      ctx = ElementContext(f['name'],'function',route.location,[])
      ctx.set_cfg(CFG(f['ast'],f['ctx'],f['cursor'],route.cfg.source_code))
      if(not self.analyze_function(ctx)):
        unsanitized_methods.append(f['name'])
    return unsanitized_methods

  def analyze_function(self,route:ElementContext)->bool:
    format_log(f"Analyzing function {route.get_base_element_name()} in {route.location}...")
    file_method_list = self.project_info.dependency_manager.get(route.location).method_list
    var_list = []
    # Cek parent nya untuk menentukaan apakah ada dekorator
    route.cfg.reset()
    node:Node = route.cfg.traverse()
    if(node.parent.type=='decorated_definition'):
      # Cari fungsi dekorator yang dipanggil
      for dec in node.parent.children:
        if dec.type=='decorator':
          if dec.children[-1].type=='call':
            dekorator = dec.children[-1].children[0]  
          else:
            dekorator = dec.children[-1] 
      # Cek apakah ada di fungsi atau library
          if self.is_in_acl_list(route,dekorator.text.decode().split("(")[0].replace("@","").strip()):
            return True
    # Gak ada dekorator/dekorator gak cocok
    # Traverse fungsinya
    while node:
      # Cek detail untuk beberapa tipe node
      if node.type in ['call','assignment']:
        # Cek call method
        # Format: [attribute,argument_list] | [identifier,argument_list]
        child_list = flatten_node(node.children)
        fun_name = ""
        for i in range(len(child_list)):
          if child_list[i] in [':',',']:
            continue
          elif child_list[i] in [')']:
              fun_name = ""
          elif child_list[i] == '=':
              # Simpan nama variabel di left hand side
              var_list.append({'name':fun_name,'type':child_list[i+1]})
              fun_name = ""
          elif child_list[i] in ['(']:
            # Cek apakah fungsinya ada
            if fun_name:
              # cek buat fungsi di kelas yang sama
              for rute in file_method_list:
                function_name = ""
                if fun_name.split(".")[0]!='self' and rute['type']=='module_function' and rute['name']==fun_name:
                  function_name = fun_name
                elif rute['type']=='class_method' and rute['name']==fun_name.split(".")[-1]:
                  if fun_name.split(".")[0]!='self':
                    function_name = fun_name
                  else:
                    function_name = fun_name.split(".")[-1]
                else:
                  function_name = fun_name
                # Cek
                if function_name and self.is_in_acl_list(route,function_name):
                  return True
              fun_name = ""
          else:
            fun_name += child_list[i]
      elif node.type=='while_statement':
        for elmt in node.children:
          # Skip kalau dah ':'
          if elmt.type==':':
            break
          elif elmt.type=='while':
            continue
          else:
            # Baca kondisi
            child_list = flatten_node(node.children)
            fun_name = ""
            for child in child_list:
              if child in [':',',']:
                continue
              elif child in ['=',')']:
                fun_name = ""
              elif child in ['('] and fun_name:
                # Cek apakah fungsinya ada
                if fun_name:
                  # cek buat fungsi di kelas yang sama
                  for rute in file_method_list:
                    function_name = ""
                    if fun_name.split(".")[0]!='self' and rute['type']=='module_function' and rute['name']==fun_name:
                      function_name = fun_name
                    elif rute['type']=='class_method' and rute['name']==fun_name.split(".")[-1]:
                      if fun_name.split(".")[0]!='self':
                        function_name = fun_name
                      else:
                        function_name = fun_name.split(".")[-1]
                    # Cek
                    if function_name and self.is_in_acl_list(route,function_name):
                      return True
                  fun_name = ""
              else:
                fun_name += child
      elif node.type=='for_statement':
        # Cek kondisinya dulu
        components = []
        for child in node.children:
          if child.type in ['for',':','in','block']:
            continue
          elif child.type=='identifier':
            components.append(child.text.decode())
          else:
            components += flatten_node(child.children)
        # Cek tiap componentnya
        child_list = flatten_node(node.children)
        fun_name = ""
        for child in child_list:
          if child in [':',',']:
            continue
          elif child in ['=',')']:
            fun_name = ""
          elif child in ['('] and fun_name:
            # Cek apakah fungsinya ada
            if fun_name:
              # cek buat fungsi di kelas yang sama
              for rute in file_method_list:
                function_name = ""
                if fun_name.split(".")[0]!='self' and rute['type']=='module_function' and rute['name']==fun_name:
                  function_name = fun_name
                elif rute['type']=='class_method' and rute['name']==fun_name.split(".")[-1]:
                  if fun_name.split(".")[0]!='self':
                    function_name = fun_name
                  else:
                    function_name = fun_name.split(".")[-1]
                # Cek
                if function_name and self.is_in_acl_list(route,function_name):
                  return True
              fun_name = ""
          else:
            fun_name += child
      elif node.type=='with_statement':
        # Cek kondisinya dulu
        components = []
        for child in node.children:
          if child.type in ['with','block',":"]:
            continue
          elif child.type=='identifier':
            components.append(child.text.decode())
          else:
            components += flatten_node(child.children)
        # Cek tiap componentnya
        fun_name = ""
        for i in range(len(components)):
          if components[i] in [':',',']:
            continue
          elif child in ['=',')']:
            fun_name = ""
          elif components[i] in ['(']:
            # Cek apakah fungsinya dipanggil
            if fun_name:
              # cek buat fungsi di kelas yang sama
              for rute in file_method_list:
                function_name = ""
                if fun_name.split(".")[0]!='self' and rute['type']=='module_function' and rute['name']==fun_name:
                  function_name = fun_name
                elif rute['type']=='class_method' and rute['name']==fun_name.split(".")[-1]:
                  if fun_name.split(".")[0]!='self':
                    function_name = fun_name
                  else:
                    function_name = fun_name.split(".")[-1]
                # Cek
                if function_name and self.is_in_acl_list(route,function_name):
                  return True
              fun_name = ""
            elif components[i]=='as':
              prefix = components[i+1]
          else:
            # Tambahkan nama fungsinya
            fun_name += components[i]
      elif node.type == 'comparison_operator':
        # Cek kondisinya dulu
        components = []
        for child in node.children:
          if child.type in ['identifier','in']:
            components.append(child.text.decode())
          elif child.type in ['string']:
            components.append(child.text.decode().replace('"','').replace("'",""))
          else:
            components += flatten_node(child.children)
        for i in range(len(components)):
          if components[i] in ['in']:
            # Cek kanan kirinya
            left_side = components[i-1]
            right_side = components[i+1]
            # Cek apakah ini acl
            if self.is_in_acl_list(route,right_side):
              return True
      elif node.type in ['try_statement','finally_clause','else_clause']:
        pass
      elif node.type=='except_clause':
        # Dapatkan identifier
        identifier = node.children[1]
      elif node.type in ['if_statement','elif_clause']:
        # Cek kondisinya dulu
        components = []
        for child in node.children:
          if child.type in ['elif',':','block']:
            continue
          elif child.type=='identifier':
            components.append(child.text.decode())
          else:
            components += flatten_node(child.children)
        # Cek tiap componentnya
        fun_name = ""
        for i in range(len(components)):
          if components[i] in ['and','or','not']:
            continue
          elif components[i] in ['(',')']:
            # Cek apakah fungsinya dipanggil
            if fun_name:
              parts = fun_name.split(".")
              # Coba konversi dulu
              if len(parts)>1:
                for var in var_list:
                  if parts[0]==var['name']:
                    parts[0] = var['type']              
              for rute in file_method_list:
                function_name = ""
                if parts[0]!='self' and rute['type']=='module_function' and rute['name']==".".join(parts):
                  function_name = ".".join(parts)
                elif rute['type']=='class_method' and rute['name']==parts[-1]:
                  if parts[0]!='self':
                    function_name = ".".join(parts)
                  else:
                    function_name = parts[-1]
                else: 
                  function_name = ".".join(parts)
                # Cek
                if function_name and self.is_in_acl_list(route,function_name,components[i+1:]):
                  return True
              fun_name = ""
          elif components[i] in ['==','!=','>','<','>=','<=']:
            # Cek operator
            continue
          else:
            # Tambahkan nama fungsinya
            fun_name += components[i]
        if fun_name:
          # Cek jika ini nama library yg diimpor
          parts = fun_name.split(".")
      node = route.cfg.traverse()
    return False

  def analyze_module(self,route:ElementContext)->list[str]:
    format_log(f"Analyzing module {route.get_base_element_name()} in {route.location}...")
    file_method_list = self.project_info.dependency_manager.get(route.location).method_list
    # Cari info semua fungsi di module
    method_lists = [m for m in file_method_list if not (m['type']=='class_method' and m['name']=='__init__')]
    # Remove yang ada di no check list
    for ctx in route.context:
      if ctx.startswith("Exempt::"):
        exempted = ctx.split("::")[1]
        for met in method_lists:
          if(met['name']==exempted or (met['type']=='class_method' and (met['parent']==exempted))):
            method_lists = [m for m in method_lists if m['name']!=met['name']]
    # Cek setiap fungsi
    unsanitized_methods = []
    for f in method_lists:
      ctx = ElementContext(f['name'],'function',route.location,[])
      # set cfg
      ctx.set_cfg(CFG(f['ast'],f['ctx'],f['cursor'],route.cfg.source_code))
      if(not self.analyze_function(ctx)):
        if f['type']=='class_method':
          unsanitized_methods.append(["class",f['name'],f['parent']])
        else:
          unsanitized_methods.append(["function",f['name']])
    return unsanitized_methods
  
  def is_in_acl_list(self,route:ElementContext,name:str,ctx:list[str]=[]):
    # Format name: fun atau self.fun
    file_method_list = self.project_info.dependency_manager.get(route.location).method_list
    for acl_class in self.project_info.acl_class:
      acl_class.cfg.reset()
      if(acl_class.type=='library' and acl_class.location==route.location):
        if(acl_class.context):
          # Cek apakah fungsi yang merupakan acl digunakan
          # context berisi daftar fungsi yang merupakan fungsi untuk cek ACL
          while True:
            acl_node:Node|None = acl_class.cfg.traverse()
            if not acl_node:
              break
            # Handle kasus antara gak ada context fungsi mana yang acl maupun ada
            if(not acl_class.context or  (acl_node.text.decode().split(".")[0] in acl_class.context)):
              # Cek jika ini fungsi ataupun kelas
              if(name==acl_class.cfg.get_name(acl_node) or name.split(".")[0]==acl_class.cfg.get_name(acl_node)):
              # Cek apakah dia manggil fungsi yang diimport di acl
                return True
      # Kasus kalau acl nya kelas atau fungsi
      elif acl_class.type in ['class','function']:
        # Cek apakah ada parent context
        parent_class = ""
        for k in acl_class.context:
          if k.startswith("Parent::"):
            parent_class = k.split("::")[1]
        if acl_class.location==route.location:
          # Kelas/fungsi lokal
          # Kali aja ada di module ini
          for function in file_method_list:
            if(acl_class.type=='class' and function['type']=='class_method' and function['parent']==acl_class.get_base_element_name()):
              full_name = name
              if len(name.split("."))==1 and ctx:
                j = 0
                elmt = ctx[j]
                while elmt in ['(',')',',']:
                  j+=1
                  if j >= len(ctx):
                    break
                  elmt = ctx[j]
                full_name = f"{name}.{elmt}"
              if(full_name==f"{function['parent']}.{function['name']}"):
                # Cek apakah dia manggil fungsi yang diimport di acl
                return True
            elif(acl_class.type=='function' and function['type']=='module_function' and function['name']==acl_class.get_base_element_name()):
              # Cukup cek apakah kelas atau fungsinya sama
              if(name==function['name']):
              # Cek apakah dia manggil fungsi yang diimport di acl
                return True
            elif(acl_class.type=='function' and function['type']=='class_method' and function['parent']==parent_class and function['name']==acl_class.get_base_element_name()):
              # Cukup cek apakah kelas atau fungsinya sama
              if(name==f'{parent_class}.{function["name"]}'):
              # Cek apakah dia manggil fungsi yang diimport di acl
                return True
        else:
          # Beda file, cek di import ada gak
          # Kali aja ada di module ini
          dependency_lists = self.project_info.dependency_manager.get(route.location).dependency_list
          route_position =  route.cfg.get_line_index(route.name)
          for dependency in dependency_lists:
            if dependency[0 ]> route_position:
              break
            dependency_parts = dependency[1]['original'] # Module name
            if dependency_parts not in acl_class.location.replace(PATH_SEPARATOR,"."):
              # Modul nya salah
              continue
            # Yang diimport class nya
            if(acl_class.type=='class'):
              # Cek apakah komponen selanjutnya method
              # Cukup cek apakah kelas atau fungsinya sama
              for m in dependency[2]:
                  comparator = m['original']
                  if m["rename"]:
                    # Pake rename
                    comparator = m["rename"]
                  if len(name.split("."))==1 and ctx:
                    j = 0
                    elmt = ctx[j]
                    while elmt in ['(',')',',']:
                      j+=1
                      if j >= len(ctx):
                        break
                      elmt = ctx[j]
                    name = f"{name}.{elmt}"
                  acl_file_method_list = self.project_info.dependency_manager.get(acl_class.location).method_list
                  for method in acl_file_method_list:
                    if method['type']!='class_method' or (method['type']=='class_method' and method['parent']!=m['original']):
                      continue
                    if(name==f'{comparator}.{method["name"]}'):
                      # Cek apakah dia manggil fungsi yang diimport di acl
                      return True
            elif(acl_class.type=='function'):
              # Cukup cek apakah kelas atau fungsinya sama
                for m in dependency[2]:
                  comparator = m['original']
                  if m["rename"]:
                    # Pake rename
                    comparator = m["rename"]
                  acl_file_method_list = self.project_info.dependency_manager.get(acl_class.location).method_list
                  for method in acl_file_method_list:
                    if method['type']=='class_method' and (f"{method['parent']}.{method['name']}"==f"{comparator}.{name}" or f"{parent_class}.{method['name']}"==f"{comparator}.{name}"):
                      return True
                    # Fungsi tidal sesuai
                    elif method['type']!='module_function' or method['name']!=m['original']:
                      continue
                    if(name==comparator):
                      # Cek apakah dia manggil fungsi yang diimport di acl
                      return True
    return False