from lib.utils import format_log
from lib.ACReader import ACReader,ACLData
from lib.FileReader import FileReader
from lib.HandlerSanitizationAnalyzer import HandlerSanitizationAnalyzer
from lib.ACLAnalyzer import ACLAnalyzer
from lib.VulnerabilitReporter import VulnerabilityReporter
import time

# Supaya support up arrow di linux, perlu import readlines
from sys import platform
if platform=='linux':
  import readline

class MainMenu():
    def __init__(self):
      self.project_path:str = ""
      self.acl_path:str = ""

    def start(self):
      self.print_header()
      # Dapatkan Path Menuju File dan Konfigurasi ACL
      is_correct:bool = False
      while not is_correct:
        self.project_path = input("Enter the path to the project: ")
        self.acl_path = input("Enter the path to the ACL file: ")
        is_correct = input("Is this correct? (y/n): ").lower() == "y"

      # Mulai analisis
      format_log("Starting Analysis...")
      generation_start_time = time.time() 
      generation_end_time = generation_start_time
      detection_start_time = generation_start_time
      detection_end_time = generation_start_time
      # Baca file ACL
      try:
        self.acl_data = ACReader(self.acl_path).read()
        format_log("ACL data acquired.")
        self.project_ctx = FileReader(self.project_path).analyze_project()
        format_log("ACL and routes context gathered...")
      except FileNotFoundError:
        format_log("File not found. Exiting...",status='error')
        generation_end_time = time.time() 
      else:
        # Inisialisasi Reporter
        reporter = VulnerabilityReporter(self.acl_data)
        generation_end_time = time.time() 
        # Analisis sanitasi router
        detection_start_time = time.time()
        format_log("Starting route sanitization analysis...")
        reporter.add_unsanitized_routes(HandlerSanitizationAnalyzer().set_project(self.project_ctx).analyze())
        # Analisis kontrol akses
        format_log("Starting access control analysis...")
        reporter.add_acl_data(ACLAnalyzer(self.project_ctx,self.acl_data).analyze())
        detection_end_time = time.time()
        # Tampilkan hasil analisis
        print("=====================ANALYSIS RESULT=====================")
        reporter.get_unsanitized_element_report()
        reporter.get_acl_table()
      finally:
        detection_time = round(detection_end_time-detection_start_time,6)
        generation_time = round(generation_end_time-generation_start_time,6)
        format_log(f"Analysis finished in {round(time.time()-generation_start_time,6)} seconds. (generation: {generation_time} seconds, detection: {detection_time} seconds)",end="\n\n")

    def print_header(self):
      print("Welcome to SCA Tool")
      print("V.1.0")

