import datetime
from tree_sitter import Node
import sys
from termcolor import colored


def format_log(message: str,*,end:str="\n",status:str='normal') -> str:
  if status =='error':
    print(colored(f"[{datetime.datetime.now()}] {message}",'red'),end=end,file=sys.stderr)    
  else:  
    print(f"[{datetime.datetime.now()}] {message}",end=end)


def flatten_node(nodes: list[Node]) -> list[str]:
  res:list[bytes] = []
  tmp_str = b""
  for node in nodes:
    if node.type in ['attribute','call','boolean_operator','not_operator','with_item','as_pattern','as_pattern_target','comparison_operator']:
      res += flatten_node(node.children)
    elif node.type in ['identifier','.']:
      tmp_str += node.text
    elif node.type =='argument_list':
      if tmp_str:
        res.append(tmp_str)
        tmp_str = b""
      content = flatten_node(node.children) 
      res += content
    elif node.type in ['=',',','and','or','not','for','in','as','==','!=','<=','>=','<','>','+','-','*','/','//','**','%','&','|','^','<<','>>','~','+=','-=','*=','/=','//=','%=','&=','|=','^=','<<=','>>=','**=','@']:
      if tmp_str:
        res.append(tmp_str)
        tmp_str = b""
      res.append(node.text)
    elif node.type in ['(',')']:
      if tmp_str:
        res.append(tmp_str)
        tmp_str = b""
      res.append(node.text)
    else: 
      tmp_str += node.text
  if tmp_str:
    res.append(tmp_str)
  # Format bagian terakhir
  if res:
    for i in range(len(res)):
      if type(res[i])==bytes:
        res[i] = res[i].decode()
    if res[-1].startswith("."):
      res[-1] = res[-1][1:]
    if res[-1].endswith("."):
      res[-1] = res[-1][:-1]
  return res
