from flask import Flask
from .MainController import register_controllers
import os

from flask_sqlalchemy import SQLAlchemy
from flask_login import LoginManager

db = SQLAlchemy()
DB_NAME = "database.db"

def create_app():
  app = Flask(__name__)

  app.config['SECRET_KEY'] = 'hjshjhdjah kjshkjdhjs'
  app.config['SQLALCHEMY_DATABASE_URI'] = f'sqlite:///{DB_NAME}'
  db.init_app(app)

  # register services
  register_controllers(app)

  #set config
  app.config['UPLOAD_PATH'] = os.path.join(app.instance_path.replace("\\instance",""),os.getenv('UPLOAD_FOLDER'))

  from .views import views
  from .auth import auth

  app.register_blueprint(views, url_prefix='/')
  app.register_blueprint(auth, url_prefix='/')
  
  from .models import User, Note
  
  create_database(app)  
    
  login_manager = LoginManager()
  login_manager.login_view = 'auth.login'
  login_manager.init_app(app)

  @login_manager.user_loader
  def load_user(id):
      return User.query.get(int(id))

  return app

def create_database(app):
    if not os.path.exists('website/' + DB_NAME):
        db.create_all(app=app)
        print('Created Database!')