from flask import Flask,render_template,request,flash,redirect,url_for
from werkzeug.utils import secure_filename
import time
import os

from src.lib.FileParser import FileParser
from src.lib.ASTGenerator import ASTGenerator
from src.lib.CFGGenerator import CFGGenerator

ALLOWED_EXTENSION = {'zip','rar','7zip'}
def register_controllers(app:Flask):
  @app.route('/')
  def main():
    return render_template("base.html")
  
  @app.route('/analyze',methods=["POST"])
  def analyze_code():
    if "files" not in request.files:
      flash("No file received")
      return redirect('/')
    archive = request.files['files']
    if(not archive):
      flash("No file received")
      return redirect('/')
    # pastikan file nya boleh diunggah
    if(not is_file_allowed(archive.filename)):
      flash("File not supported")
      return redirect('/')
    # rename
    names = archive.filename.rsplit('.',1)
    filename = f"{secure_filename(names[0])}-{str(int(time.time()))}.{names[1]}"
    # simpan dulu    
    archive_path = os.path.join(app.config['UPLOAD_PATH'],filename)
    archive.save(archive_path)
    # parse file
    parser = FileParser(archive_path)
    files = parser.load()
    # get folder name
    folder_name = parser.folder_name
    # generate ast
    ast_generator = ASTGenerator(files)
    ast_generator.generate_ast(folder_name)
    # generate cfg
    cfg_generator = CFGGenerator()
    cfg_generator.generate_cfg(folder_name)
    return "TBD"
  
  def is_file_allowed(filename:str):
    return '.' in filename and filename.rsplit('.',1)[1].lower() in ALLOWED_EXTENSION  
  